#define _GNU_SOURCE

#include <stdio.h>

#ifdef WIN32
# include <windows.h>
# include <tlhelp32.h>
#else
# include <dlfcn.h>
// this cond is from ruby
# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || (defined(__OpenBSD__) && !defined(__ELF__)) || defined(NeXT) || defined(__WATCOMC__) || defined(__APPLE__)
# include <mach-o/dyld.h>
#else
# include <link.h>
#endif
#endif

#include <bfd.h>
#include <ffi.h>
#include "demangle.h"

#include <ruby.h>

typedef struct {
    char* full_name;
    char* name;
    ffi_type** args;
    int arg_num;
    int fp;
} reflection_sym_t;

typedef struct {
    reflection_sym_t** syms;
    int sym_num;
    int class_size;
    char* class_name;
} rubycpp;

void* _Znwj(int size);

static char* mstrncpy(const char* src, int i) {
    char* ret = (char*)malloc(sizeof(char)*(i+1));
    strncpy(ret, src, i);
    ret[i] = '\0';
    return ret;
}

static char* mstrcpy(const char* src) {
    return mstrncpy(src, strlen(src));
}

static VALUE rubycpp_invoke(rubycpp* cpp, void* vp, char* mn,
                            int argc, VALUE* argv)
{
    int i;
    int* as;
    void** args;
    as = (int*)malloc(sizeof(int)*argc);
    argc++;
    args = (void**)malloc(sizeof(void*)*argc);
    args[0] = &vp;
    for (i = 0; i < argc-1; i++) {
        switch (TYPE(argv[i])) {
        case T_FIXNUM:
            as[i] = NUM2INT(argv[i]); break;
        case T_STRING:
            as[i] = StringValuePtr(argv[i]); break;
        }
        args[i+1] = &as[i];
    }
    for (i = 0; i < cpp->sym_num; i++) {
        reflection_sym_t* s = cpp->syms[i];
        if (strcmp(s->name, mn) == 0 && argc == s->arg_num) {
            ffi_cif cif;
            ffi_type* rt = &ffi_type_sint32;
            ffi_arg rv;
            printf("[RUBY] invoke %s\n", s->full_name);
            if (ffi_prep_cif(&cif, FFI_DEFAULT_ABI, argc, rt, s->args) !=
                FFI_OK)
            {
                printf("err\n");
                return Qnil;
            }
            ffi_call(&cif, FFI_FN(s->fp), &rv, args);
            free(as);
            free(args);
            return INT2NUM(rv);
        }
    }
    printf("error\n");
    return Qnil;
}

static VALUE rubycpp_initialize(int argc, VALUE* argv, VALUE obj) {
    void* vp;
    rubycpp* cpp;
    VALUE klass;
    char buf[256];

    klass = rb_obj_class(obj);
    cpp = (rubycpp*)rb_cvar_get(klass, 0);

    strcpy(buf, cpp->class_name);
    strcat(buf, "::");
    strcat(buf, cpp->class_name);

    vp = _Znwj(cpp->class_size);
    rubycpp_invoke(cpp, vp, buf, argc, argv);
    rb_ivar_set(obj, 0, (VALUE)vp);

    return obj;
}

static VALUE rubycpp_method_missing(int argc, VALUE* argv, VALUE obj) {
    void* vp;
    rubycpp* cpp;
    VALUE klass;
    char buf[256];
    char* mn;

    klass = rb_obj_class(obj);
    cpp = (rubycpp*)rb_cvar_get(klass, 0);

    strcpy(buf, cpp->class_name);
    strcat(buf, "::");
    strcat(buf, rb_id2name(SYM2ID(argv[0])));

    vp = rb_ivar_get(obj, 0);

    return rubycpp_invoke(cpp, vp, buf, argc-1, argv+1);
}

static reflection_sym_t* init_sym() {
    reflection_sym_t* r = (reflection_sym_t*)malloc(sizeof(reflection_sym_t));
    r->full_name = NULL;
    r->name = NULL;
    r->args = NULL;
    r->arg_num = 0;
    r->fp = 0;
    return r;
}

static void delete_sym(reflection_sym_t* r) {
    int i;
    free(r->full_name);
    free(r->name);
    free(r->args);
    free(r);
}

static ffi_type* parse_type(const char* str) {
    char* p;
    if (strcmp(str, "int") == 0) return &ffi_type_sint32;
    if (strcmp(str, "char") == 0) return &ffi_type_sint8;
    if (strcmp(str, "float") == 0) return &ffi_type_float;
/*    if (strcmp(str, "double") == 0) return &ffi_type_double; */
    p = str+strlen(str)-1;
    if (*p == '*') return &ffi_type_pointer;
    return NULL;
}

static reflection_sym_t* parse_sym_signature(reflection_sym_t* sym) {
    char* pb;
    char* pe;

    pb = index(sym->full_name, '(');
    if (pb == NULL) {
        delete_sym(sym);
        return sym;
    }
    sym->name = mstrncpy(sym->full_name, pb - sym->full_name);

    sym->arg_num = 1;
    sym->args = (ffi_type**)malloc(sizeof(ffi_type*));
    sym->args[0] = &ffi_type_pointer;

    pb++;
    pe = pb;
    while (*pe != ')') {
        pe++;
        if (*pe == '(') {
            while (*pe != ')') pe++;
            pe++;
        }
        if (*pe == '<') {
            while (*pe != '>') pe++;
            pe++;
        }
        if (*pe == ',' || *pe == ')') {
            char buf[256];
            strncpy(buf, pb, pe-pb);
            buf[pe-pb] = '\0';
            sym->arg_num++;
            sym->args = (ffi_type**)realloc(
                sym->args, sizeof(ffi_type*)*sym->arg_num);
            sym->args[sym->arg_num-1] = parse_type(buf);
//            if (!sym->args[sym->arg_num-1]) printf("%s %s\n", buf, sym->full_name);
            pb = pe + 2;
        }
    }

    return sym;
}

static void add_symbol_impl(rubycpp* syms, const char* name, int value) {
    reflection_sym_t* sym;

    sym = init_sym();
    sym->full_name =
        cplus_demangle(name, DMGL_PARAMS|DMGL_ANSI|DMGL_VERBOSE|DMGL_TYPES);
    sym->fp = value;

    sym = parse_sym_signature(sym);
    if (!sym) return;

    syms->sym_num++;
    syms->syms = (reflection_sym_t**)
        realloc(syms->syms, sizeof(reflection_sym_t*)*syms->sym_num);
    syms->syms[syms->sym_num-1] = sym;

//    printf("%08x %s %s %d\n", sym->fp, sym->name, sym->full_name, sym->arg_num);
}

static void add_symbol(rubycpp* syms, int offset, asymbol* asym) {
    const char* name;
    int value;

    name = bfd_asymbol_name(asym);
    value = bfd_asymbol_value(asym);

    if (name == NULL || strlen(name) < 2) return;
    if (value == 0) return;

#ifdef WIN32
    if (name[0] != '_') return;
    name++;
#endif

#ifdef __APPLE__
    if (name[0] != '_') return;
    name++;
#endif

    if (name[0] != '_' || name[1] != 'Z') return;

    if ((unsigned int)value < (unsigned int)offset) {
        value += offset;
    }

    add_symbol_impl(syms, name, value);
}

static rubycpp* read_symbols(const char* filename, int offset) {
    rubycpp* ret;
    bfd* abfd;
    int i;
    int symnum;

//    printf("reading %s(0x%x)... \n", filename, offset);

    abfd = bfd_openr(filename, NULL);
    if (abfd == NULL) {
        bfd_perror("bfd_openr");
        return 1;
    }
    if (!bfd_check_format(abfd, bfd_object)) {
        bfd_perror("bfd_check_format");
        return 1;
    }

    {
        void* minisyms;
        unsigned int size;
        char* p;
        asymbol* store;
        int dyn = 0;

        store = bfd_make_empty_symbol(abfd);

        symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
#if defined(__linux__) || defined(__FreeBSD__)
        if (symnum == 0) {
            dyn = 1;
            symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
        }
#endif
        if (symnum < 0) {
            bfd_perror("bfd_read_minisymbols");
            goto error;
        }

        ret = (rubycpp*)malloc(sizeof(rubycpp));
        ret->syms = NULL;
        ret->sym_num = 0;

        p = (char*)minisyms;
        for (i = 0; i < symnum; i++) {
            asymbol* sym;
            sym = bfd_minisymbol_to_symbol(abfd, dyn, p, store);
            add_symbol(ret, offset, sym);
            p += size;
        }

        free(minisyms);
    }

    bfd_close(abfd);
    return ret;

error:
    bfd_close(abfd);
    return 1;
}

static rubycpp* make_syms(const char* so) {
    const char* file = so;
    void* dlhandle;
    char buf[256];

    strcpy(buf, "lib");
    strcat(buf, file);
#ifdef __APPLE__
    strcat(buf, ".dylib");
#else
    strcat(buf, ".so");
#endif
    file = buf;

    if (!(dlhandle = dlopen(file, RTLD_NOW))) {
        char b[256];
        strcpy(b, "./");
        strcat(b, file);
        if (!(dlhandle = dlopen(b, RTLD_NOW))) {
            perror("cannot dlopen");
        }
        /* @@@ error handling */
    }

/*#ifdef HAVE_DLINFO*/
    {
        struct link_map* lmap;

        if (dlinfo(dlhandle, RTLD_DI_LINKMAP, &lmap) == 0) {
            while (lmap) {
                if (strstr(lmap->l_name, so)) {
                    return read_symbols(lmap->l_name, lmap->l_addr);
                }
                lmap = lmap->l_next;
            }
            /* @@@ error handling */
        }
        else {
            perror("dlinfo");
            /* @@@ error handling */
        }
    }
    return NULL;
}

static VALUE rubycpp_declare_class(
    VALUE self, VALUE sov, VALUE cnv, VALUE size)
{
    VALUE cpp;
    char* so;
    char* rbname;
    rubycpp* rubycpp;
    VALUE syms;

    StringValue(sov);
    StringValue(cnv);

    so = StringValuePtr(sov);

    rbname =  (char*)malloc(strlen(so)+12);
    strcpy(rbname, "__rubycpp__");
    strcat(rbname, so);
    cpp = rb_define_class(rbname, rb_cObject);
    free(rbname);

    rubycpp = make_syms(so);
    rubycpp->class_size = NUM2INT(size);
    rubycpp->class_name = mstrcpy(StringValuePtr(cnv));
    syms = (VALUE)rubycpp;
    rb_cvar_set(cpp, 0, syms, Qtrue);
    rb_define_method(cpp, "initialize", rubycpp_initialize, -1);
    rb_define_method(cpp, "method_missing", rubycpp_method_missing, -1);

    return cpp;
}

void Init_rubycpp() {
    VALUE cplusplus = rb_define_module("Cplusplus");
    VALUE dispatch = rb_define_class_under(cplusplus, "Dispatch", rb_cObject);
    rb_define_singleton_method(cplusplus, "declare_class",
                               rubycpp_declare_class, 3);
}

